import {
	Callable,
	type Constructor,
	type Entry,
	type Guardable,
	type Json,
	type JsonData,
	type conform,
	type evaluate
} from "@arktype/util"
import type { PredicateNode } from "./constraints/predicate.js"
import type { IndexNode } from "./constraints/props/index.js"
import type { PropNode } from "./constraints/props/prop.js"
import type { SequenceNode } from "./constraints/props/sequence.js"
import type { DivisorNode } from "./constraints/refinements/divisor.js"
import type { BoundNodesByKind } from "./constraints/refinements/kinds.js"
import type { RegexNode } from "./constraints/refinements/regex.js"
import type {
	Declaration,
	Inner,
	Schema,
	ioKindOf,
	reducibleKindOf
} from "./kinds.js"
import type { Resolutions, Scope } from "./scope.js"
import type { NodeCompiler } from "./shared/compile.js"
import type {
	BaseAttachmentsOf,
	BaseErrorContext,
	BaseMeta,
	BaseNodeDeclaration
} from "./shared/declare.js"
import type { ArkResult } from "./shared/errors.js"
import {
	precedenceOfKind,
	type BasisKind,
	type ConstraintKind,
	type NodeKind,
	type PropKind,
	type RefinementKind,
	type TypeKind,
	type UnknownIntersectionResult,
	type UnknownNodeImplementation,
	type nodeImplementationInputOf,
	type nodeImplementationOf
} from "./shared/implement.js"
import { inferred } from "./shared/inference.js"
import { type TraverseAllows, type TraverseApply } from "./shared/traversal.js"
import type { DomainNode } from "./types/domain.js"
import type { IntersectionNode } from "./types/intersection.js"
import type {
	MorphNode,
	distillConstrainableIn,
	distillConstrainableOut,
	distillIn,
	distillOut
} from "./types/morph.js"
import type { ProtoNode } from "./types/proto.js"
import type { Type } from "./types/type.js"
import type { UnionNode } from "./types/union.js"
import type { UnitNode } from "./types/unit.js"
import { arkKind } from "./util.js"
export interface BaseAttachments {
	alias?: string
	readonly kind: NodeKind
	readonly reference: string
	readonly inner: Record<string, any>
	readonly entries: readonly Entry<string>[]
	readonly json: object
	readonly typeJson: object
	readonly collapsibleJson: JsonData
	readonly children: UnknownNode[]
	readonly innerId: string
	readonly typeId: string
	readonly $: Scope
}
export interface NarrowedAttachments<d extends BaseNodeDeclaration>
	extends BaseAttachments {
	kind: d["kind"]
	inner: d["inner"]
	json: Json
	typeJson: Json
	collapsibleJson: JsonData
	children: Node<d["childKind"]>[]
}
export type NodeSubclass<d extends BaseNodeDeclaration = BaseNodeDeclaration> =
	{
		new (attachments: never): BaseNode<any, d>
		readonly implementation: nodeImplementationOf<d>
	}
export type UnknownNode = BaseNode<any, BaseNodeDeclaration>
type subclassKind<self> = self extends Constructor<{
	kind: infer kind extends NodeKind
}>
	? kind
	: never
type subclassDeclaration<self> = Declaration<subclassKind<self>>
export declare abstract class BaseNode<
	t,
	d extends BaseNodeDeclaration
> extends Callable<
	(data: d["prerequisite"]) => ArkResult<distillIn<t>, distillOut<t>>,
	BaseAttachmentsOf<d>
> {
	attachments: BaseAttachments
	constructor(attachments: BaseAttachments)
	infer: distillOut<t>;
	[inferred]: t
	readonly [arkKind] = "node"
	protected static implement<self>(
		this: self,
		implementation: nodeImplementationInputOf<subclassDeclaration<self>>
	): nodeImplementationOf<subclassDeclaration<self>>
	protected readonly impl: UnknownNodeImplementation
	readonly includesMorph: boolean
	readonly includesContextDependentPredicate: boolean
	readonly referencesByName: Record<string, UnknownNode>
	readonly references: readonly UnknownNode[]
	readonly contributesReferencesByName: Record<string, UnknownNode>
	readonly contributesReferences: readonly UnknownNode[]
	readonly precedence: precedenceOfKind<d["kind"]>
	jit: boolean
	private descriptionCache?
	get description(): string
	private compiledErrorContextCache
	get compiledErrorContext(): string
	allows: (data: d["prerequisite"]) => data is this["in"]["infer"]
	parse(
		data: d["prerequisite"]
	): ArkResult<this["in"]["infer"], this["infer"]>
	abstract traverseAllows: TraverseAllows<d["prerequisite"]>
	abstract traverseApply: TraverseApply<d["prerequisite"]>
	abstract compile(js: NodeCompiler): void
	abstract expression: string
	private inCache?
	get in(): Node<ioKindOf<d["kind"]>, distillConstrainableIn<t>>
	private outCache?
	get out(): Node<ioKindOf<d["kind"]>, distillConstrainableOut<t>>
	private getIo
	protected createErrorContext<from>(
		from: from
	): evaluate<BaseErrorContext<d["kind"]> & from>
	toJSON(): Json
	equals(other: UnknownNode): boolean
	hasKind<kind extends NodeKind>(kind: kind): this is Node<kind>
	isBasis(): this is Node<BasisKind>
	isConstraint(): this is ConstraintNode
	isRefinement(): this is Node<RefinementKind>
	isProp(): this is Node<PropKind>
	isType(): this is Type
	hasUnit<value>(value: unknown): this is UnitNode<value>
	get intersectionIsOpen(): d["intersectionIsOpen"]
	get nestableExpression(): string
	private static intersectionCache
	protected intersectInternal(
		this: UnknownNode,
		r: UnknownNode
	): UnknownIntersectionResult
	firstReference<narrowed>(
		filter: Guardable<UnknownNode, conform<narrowed, UnknownNode>>
	): narrowed | undefined
	firstReferenceOrThrow<narrowed extends UnknownNode>(
		filter: Guardable<UnknownNode, narrowed>
	): narrowed
	firstReferenceOfKind<kind extends NodeKind>(
		kind: kind
	): Node<kind> | undefined
	firstReferenceOfKindOrThrow<kind extends NodeKind>(kind: kind): Node<kind>
	bindScope<resolutions extends Resolutions>(
		$: Scope<resolutions>
	): Node<d["kind"], t, resolutions>
	transform(
		mapper: DeepNodeTransformation,
		shouldTransform: (node: UnknownNode) => boolean
	): Node<reducibleKindOf<this["kind"]>>
	configureShallowDescendants(configOrDescription: BaseMeta | string): this
}
export type DeepNodeTransformation = <kind extends NodeKind>(
	kind: kind,
	inner: Inner<kind>
) => Inner<kind>
interface NodesByKind<t = any, $ = any> extends BoundNodesByKind {
	union: UnionNode<t, $>
	morph: MorphNode<t, $>
	intersection: IntersectionNode<t, $>
	unit: UnitNode<t, $>
	proto: ProtoNode<t, $>
	domain: DomainNode<t, $>
	divisor: DivisorNode
	regex: RegexNode
	predicate: PredicateNode
	prop: PropNode
	index: IndexNode
	sequence: SequenceNode
}
export type Node<kind extends NodeKind, t = any, $ = any> = NodesByKind<
	t,
	$
>[kind]
export type TypeSchema<kind extends TypeKind = TypeKind> = Schema<kind>
export declare const typeKindOfSchema: (schema: unknown) => TypeKind
export type ConstraintNode<kind extends ConstraintKind = ConstraintKind> =
	Node<kind>
export {}
//# sourceMappingURL=base.d.ts.map
