import {
	Callable,
	compileSerializedValue,
	flatMorph,
	includes,
	isArray,
	printable,
	throwError,
	throwParseError
} from "@arktype/util"
import { Disjoint } from "./shared/disjoint.js"
import {
	basisKinds,
	constraintKinds,
	discriminatingIntersectionKeys,
	precedenceOfKind,
	propKinds,
	refinementKinds,
	typeKinds
} from "./shared/implement.js"
import { inferred } from "./shared/inference.js"
import { TraversalContext } from "./shared/traversal.js"
import { arkKind, hasArkKind } from "./util.js"
export class BaseNode extends Callable {
	attachments
	constructor(attachments) {
		super(
			(data) => {
				if (
					!this.includesMorph &&
					!this.includesContextDependentPredicate &&
					this.allows(data)
				) {
					return { data, out: data }
				}
				const ctx = new TraversalContext(data, this.$.resolvedConfig)
				this.traverseApply(data, ctx)
				return ctx.finalize()
			},
			{ attach: attachments }
		)
		this.attachments = attachments
		this.contributesReferencesByName =
			this.reference in this.referencesByName
				? this.referencesByName
				: { ...this.referencesByName, [this.reference]: this }
		this.contributesReferences = Object.values(
			this.contributesReferencesByName
		)
	}
	[arkKind] = "node"
	static implement(_) {
		const implementation = _
		if (implementation.hasAssociatedError) {
			implementation.defaults.expected ??= (ctx) =>
				"description" in ctx
					? ctx.description
					: // TODO: does passing ctx here work? or will some expect node?
						implementation.defaults.description(ctx)
			implementation.defaults.actual ??= (data) => printable(data)
			implementation.defaults.problem ??= (ctx) =>
				`must be ${ctx.expected}${
					ctx.actual ? ` (was ${ctx.actual})` : ""
				}`
			implementation.defaults.message ??= (ctx) => {
				if (ctx.path.length === 0) return ctx.problem
				const problemWithLocation = `${ctx.propString} ${ctx.problem}`
				if (problemWithLocation[0] === "[") {
					// clarify paths like [1], [0][1], and ["key!"] that could be confusing
					return `value at ${problemWithLocation}`
				}
				return problemWithLocation
			}
		}
		return implementation
	}
	impl = this.constructor.implementation
	includesMorph =
		this.kind === "morph" ||
		this.children.some((child) => child.includesMorph)
	includesContextDependentPredicate =
		// if a predicate accepts exactly one arg, we can safely skip passing context
		(this.hasKind("predicate") && this.inner.predicate.length !== 1) ||
		this.children.some((child) => child.includesContextDependentPredicate)
	referencesByName = this.children.reduce(
		(result, child) =>
			Object.assign(result, child.contributesReferencesByName),
		{}
	)
	references = Object.values(this.referencesByName)
	contributesReferencesByName
	contributesReferences
	precedence = precedenceOfKind(this.kind)
	jit = false
	descriptionCache
	get description() {
		this.descriptionCache ??=
			this.inner.description ??
			this.$.resolvedConfig[this.kind].description?.(this)
		return this.descriptionCache
	}
	compiledErrorContextCache
	get compiledErrorContext() {
		if (!this.compiledErrorContextCache) {
			if ("errorContext" in this) {
				let result = "{ "
				for (const [k, v] of Object.entries(this.errorContext)) {
					result += `${k}: ${compileSerializedValue(v)}, `
				}
				this.compiledErrorContextCache = result + " }"
			} else {
				this.compiledErrorContextCache = "{}"
			}
		}
		return this.compiledErrorContextCache
	}
	allows = (data) => {
		const ctx = new TraversalContext(data, this.$.resolvedConfig)
		return this.traverseAllows(data, ctx)
	}
	parse(data) {
		return this(data)
	}
	inCache
	get in() {
		this.inCache ??= this.getIo("in")
		return this.inCache
	}
	outCache
	get out() {
		this.outCache ??= this.getIo("out")
		return this.outCache
	}
	getIo(kind) {
		if (!this.includesMorph) {
			return this
		}
		const ioInner = {}
		for (const [k, v] of this.entries) {
			const keyDefinition = this.impl.keys[k]
			if (keyDefinition.meta) {
				continue
			}
			if (keyDefinition.child) {
				const childValue = v
				ioInner[k] = isArray(childValue)
					? childValue.map((child) => child[kind])
					: childValue[kind]
			} else {
				ioInner[k] = v
			}
		}
		return this.$.node(this.kind, ioInner)
	}
	createErrorContext(from) {
		return Object.freeze({
			code: this.kind,
			description: this.description,
			...from
		})
	}
	toJSON() {
		return this.json
	}
	equals(other) {
		return this.typeId === other.typeId
	}
	hasKind(kind) {
		return this.kind === kind
	}
	isBasis() {
		return includes(basisKinds, this.kind)
	}
	isConstraint() {
		return includes(constraintKinds, this.kind)
	}
	isRefinement() {
		return includes(refinementKinds, this.kind)
	}
	isProp() {
		return includes(propKinds, this.kind)
	}
	isType() {
		return includes(typeKinds, this.kind)
	}
	hasUnit(value) {
		return this.hasKind("unit") && this.allows(value)
	}
	get intersectionIsOpen() {
		return this.impl.intersectionIsOpen
	}
	get nestableExpression() {
		return this.children.length > 1 &&
			this.children.some((child) => !child.isBasis && !child.isProp())
			? `(${this.expression})`
			: this.expression
	}
	static intersectionCache = {}
	intersectInternal(r) {
		const lrCacheKey = `${this.typeId}&${r.typeId}`
		if (BaseNode.intersectionCache[lrCacheKey]) {
			return BaseNode.intersectionCache[lrCacheKey]
		}
		const rlCacheKey = `${r.typeId}&${this.typeId}`
		if (BaseNode.intersectionCache[rlCacheKey]) {
			// if the cached result was a Disjoint and the operands originally
			// appeared in the opposite order, we need to invert it to match
			const rlResult = BaseNode.intersectionCache[rlCacheKey]
			const lrResult =
				rlResult instanceof Disjoint ? rlResult.invert() : rlResult
			// add the lr result to the cache directly to bypass this check in the future
			BaseNode.intersectionCache[lrCacheKey] = lrResult
			return lrResult
		}
		if (this.equals(r)) {
			return this
		}
		const leftmostKind = this.precedence < r.precedence ? this.kind : r.kind
		const implementation =
			this.impl.intersections[r.kind] ?? r.impl.intersections[this.kind]
		let result =
			implementation === undefined
				? // should be two ConstraintNodes that have no relation
					// this could also happen if a user directly intersects a Type and a ConstraintNode,
					// but that is not allowed by the external function signature
					null
				: leftmostKind === this.kind
					? implementation(this, r, this.$)
					: implementation(r, this, this.$)
		if (result instanceof BaseNode) {
			// if the result equals one of the operands, preserve its metadata by
			// returning the original reference
			if (this.equals(result)) result = this
			else if (r.equals(result)) result = r
		}
		BaseNode.intersectionCache[lrCacheKey] = result
		return result
	}
	firstReference(filter) {
		return this.references.find(filter)
	}
	firstReferenceOrThrow(filter) {
		return (
			this.firstReference(filter) ??
			throwError(
				`${this.reference} had no references matching predicate ${filter}`
			)
		)
	}
	firstReferenceOfKind(kind) {
		return this.firstReference((node) => node.kind === kind)
	}
	firstReferenceOfKindOrThrow(kind) {
		return (
			this.firstReference((node) => node.kind === kind) ??
			throwError(`${this.reference} had no ${kind} references`)
		)
	}
	bindScope($) {
		if (this.$ === $) return this
		return new this.constructor({ ...this.attachments, $ })
	}
	transform(mapper, shouldTransform) {
		if (!shouldTransform(this)) {
			return this
		}
		const innerWithTransformedChildren = flatMorph(this.inner, (k, v) => [
			k,
			this.impl.keys[k].child
				? isArray(v)
					? v.map((node) => node.transform(mapper, shouldTransform))
					: v.transform(mapper, shouldTransform)
				: v
		])
		return this.$.node(
			this.kind,
			mapper(this.kind, innerWithTransformedChildren)
		)
	}
	configureShallowDescendants(configOrDescription) {
		const config =
			typeof configOrDescription === "string"
				? { description: configOrDescription }
				: configOrDescription
		return this.transform(
			(kind, inner) => ({ ...inner, ...config }),
			(node) => !node.isProp()
		)
	}
}
export const typeKindOfSchema = (schema) => {
	switch (typeof schema) {
		case "string":
			return "domain"
		case "function":
			return hasArkKind(schema, "node")
				? schema.isType()
					? schema.kind
					: throwParseError(
							`${schema.kind} constraint ${schema.expression} cannot be used as a root type`
						)
				: "proto"
		case "object":
			// throw at end of function
			if (schema === null) break
			if ("morphs" in schema) return "morph"
			if ("branches" in schema || isArray(schema)) return "union"
			if ("unit" in schema) return "unit"
			const schemaKeys = Object.keys(schema)
			if (
				schemaKeys.length === 0 ||
				schemaKeys.some((k) => k in discriminatingIntersectionKeys)
			)
				return "intersection"
			if ("proto" in schema) return "proto"
			if ("domain" in schema) return "domain"
	}
	return throwParseError(`${printable(schema)} is not a valid type schema`)
}
//# sourceMappingURL=base.js.map
