import {
	entriesOf,
	hasDomain,
	isArray,
	isKeyOf,
	registry,
	throwParseError
} from "@arktype/util"
import { BaseNode } from "./base.js"
import { nodesByKind } from "./kinds.js"
import { Disjoint } from "./shared/disjoint.js"
import { defaultValueSerializer, precedenceOfKind } from "./shared/implement.js"
const typeCountsByPrefix = {}
const baseKeys = {
	description: { meta: true }
}
// eslint-disable-next-line prefer-arrow-functions/prefer-arrow-functions
export function parseAttachments(kind, schema, ctx) {
	const cls = nodesByKind[kind]
	const impl = cls.implementation
	if (schema instanceof BaseNode && schema.kind === kind) {
		return schema
	}
	const normalizedDefinition = impl.normalize?.(schema) ?? schema
	// check again after normalization in case a node is a valid collapsed
	// schema for the kind (e.g. sequence can collapse to element accepting a Node)
	if (normalizedDefinition instanceof BaseNode) {
		return normalizedDefinition.kind === kind
			? normalizedDefinition
			: throwMismatchedNodeSchemaError(kind, normalizedDefinition.kind)
	}
	const inner = {}
	// ensure node entries are parsed in order of precedence, with non-children
	// parsed first
	const schemaEntries = entriesOf(normalizedDefinition).sort(
		([lKey], [rKey]) =>
			isKeyOf(lKey, nodesByKind)
				? isKeyOf(rKey, nodesByKind)
					? precedenceOfKind(lKey) - precedenceOfKind(rKey)
					: 1
				: isKeyOf(rKey, nodesByKind)
					? -1
					: lKey < rKey
						? -1
						: 1
	)
	const children = []
	for (const entry of schemaEntries) {
		const k = entry[0]
		const keyImpl = impl.keys[k] ?? baseKeys[k]
		if (!keyImpl) {
			return throwParseError(`Key ${k} is not valid on ${kind} schema`)
		}
		const v = keyImpl.parse ? keyImpl.parse(entry[1], ctx) : entry[1]
		if (v !== undefined || keyImpl.preserveUndefined) {
			inner[k] = v
		}
	}
	const entries = entriesOf(inner)
	let json = {}
	let typeJson = {}
	let collapsibleJson = {}
	entries.forEach(([k, v]) => {
		const keyImpl = impl.keys[k] ?? baseKeys[k]
		if (keyImpl.child) {
			const listableNode = v
			if (isArray(listableNode)) {
				json[k] = listableNode.map((node) => node.collapsibleJson)
				children.push(...listableNode)
			} else {
				json[k] = listableNode.collapsibleJson
				children.push(listableNode)
			}
		} else {
			json[k] = keyImpl.serialize
				? keyImpl.serialize(v)
				: defaultValueSerializer(v)
		}
		if (!keyImpl.meta) {
			typeJson[k] = json[k]
		}
		if (!keyImpl.implied) {
			collapsibleJson[k] = json[k]
		}
	})
	// check keys on collapsibleJson instead of schema in case one or more keys is
	// implied, e.g. minVariadicLength on a SequenceNode
	const collapsibleKeys = Object.keys(collapsibleJson)
	if (
		collapsibleKeys.length === 1 &&
		collapsibleKeys[0] === impl.collapsibleKey
	) {
		collapsibleJson = collapsibleJson[impl.collapsibleKey]
		if (
			// if the collapsibleJson is still an object
			hasDomain(collapsibleJson, "object") &&
			// and the JSON did not include any implied keys
			Object.keys(json).length === 1
		) {
			// we can replace it with its collapsed value
			json = collapsibleJson
			typeJson = collapsibleJson
		}
	}
	const innerId = JSON.stringify({ kind, ...json })
	if (ctx.reduceTo) {
		return (ctx.$.nodeCache[innerId] = ctx.reduceTo)
	}
	const typeId = JSON.stringify({ kind, ...typeJson })
	if (impl.reduce && !ctx.prereduced) {
		const reduced = impl.reduce(inner, ctx.$)
		if (reduced) {
			if (reduced instanceof Disjoint) return reduced.throw()
			// if we're defining the resolution of an alias and the result is
			// reduced to another node, add the alias to that node if it doesn't
			// already have one.
			if (ctx.alias) {
				reduced.alias ??= ctx.alias
			}
			// we can't cache this reduction for now in case the reduction involved
			// impliedSiblings
			return reduced
		}
	}
	// we have to wait until after reduction to return a cached entry,
	// since reduction can add impliedSiblings
	if (innerId in ctx.$.nodeCache) {
		return ctx.$.nodeCache[innerId]
	}
	const prefix = ctx.alias ?? kind
	typeCountsByPrefix[prefix] ??= 0
	const reference = `${prefix}${++typeCountsByPrefix[prefix]}`
	const attachments = {
		reference,
		kind,
		inner,
		entries,
		json: json,
		typeJson: typeJson,
		collapsibleJson: collapsibleJson,
		children,
		innerId,
		typeId,
		$: ctx.$
	}
	if (ctx.alias) {
		attachments.alias = ctx.alias
	}
	for (const k in inner) {
		// avoid conflict with builtin cached getters
		if (k !== "description" && k !== "in" && k !== "out") {
			attachments[k] = inner[k]
		}
	}
	return (ctx.$.nodeCache[innerId] = new cls(attachments))
}
const throwMismatchedNodeSchemaError = (expected, actual) =>
	throwParseError(
		`Node of kind ${actual} is not valid as a ${expected} definition`
	)
//# sourceMappingURL=parse.js.map
