import {
	CompiledFunction,
	isArray,
	printable,
	throwInternalError,
	throwParseError
} from "@arktype/util"
import { globalConfig } from "./config.js"
import { nodesByKind } from "./kinds.js"
import { parseAttachments } from "./parse.js"
import { NodeCompiler } from "./shared/compile.js"
import { discriminatingIntersectionKeys } from "./types/intersection.js"
import { BaseType } from "./types/type.js"
const parseConfig = (scopeConfig) => {
	if (!scopeConfig) {
		return globalConfig
	}
	const parsedConfig = { ...globalConfig }
	let k
	for (k in scopeConfig) {
		if (k === "prereducedAliases") {
			parsedConfig[k] = scopeConfig[k]
		} else {
			parsedConfig[k] = {
				...nodesByKind[k].implementation.defaults,
				...scopeConfig[k]
			}
		}
	}
	return parsedConfig
}
const assertTypeKind = (schema) => {
	switch (typeof schema) {
		case "string":
			return "domain"
		case "function":
			return "proto"
		case "object":
			// throw at end of function
			if (schema === null) break
			if (schema instanceof BaseType) return schema.kind
			if ("morph" in schema) return "morph"
			if ("branches" in schema || isArray(schema)) return "union"
			if ("unit" in schema) return "unit"
			const schemaKeys = Object.keys(schema)
			if (
				schemaKeys.length === 0 ||
				schemaKeys.some((k) => k in discriminatingIntersectionKeys)
			)
				return "intersection"
			if ("proto" in schema) return "proto"
			if ("domain" in schema) return "domain"
	}
	return throwParseError(`${printable(schema)} is not a valid type schema`)
}
export class ScopeNode {
	def
	nodeCache = {}
	config
	resolutions = {}
	referencesByName = {}
	references
	resolved = false
	prereducedAliases
	lengthBoundable
	constructor(def, config = {}) {
		this.def = def
		this.config = parseConfig(config)
		this.prereducedAliases = config.prereducedAliases ?? false
		for (const k in this.def) {
			this.resolutions[k] = this.parseRoot(
				assertTypeKind(this.def[k]),
				this.def[k],
				{
					alias: k,
					prereduced: this.prereducedAliases
				}
			)
		}
		this.references = Object.values(this.referencesByName)
		this.bindCompiledScope(this.references)
		this.resolved = true
		this.parse(
			"union",
			{
				branches: [
					"string",
					"number",
					"object",
					"bigint",
					"symbol",
					{ unit: true },
					{ unit: false },
					{ unit: null },
					{ unit: undefined }
				]
			},
			{ reduceTo: this.parsePrereduced("intersection", {}) }
		)
		this.lengthBoundable = this.parsePrereduced("union", ["string", Array])
	}
	get tsKeywords() {
		return ScopeNode.tsKeywords
	}
	get jsObjects() {
		return ScopeNode.jsObjects
	}
	static from(aliases, config = {}) {
		return new ScopeNode(aliases, config)
	}
	static root = this.from({})
	parseUnion(input) {
		return this.parseRoot("union", input)
	}
	parseBranches(...branches) {
		return branches.length === 1
			? this.parseTypeNode(branches[0], { root: true })
			: this.parseRoot("union", branches)
	}
	parseUnits(...values) {
		const uniqueValues = []
		for (const value of values) {
			if (!uniqueValues.includes(value)) {
				uniqueValues.push(value)
			}
		}
		const branches = uniqueValues.map((unit) =>
			this.parsePrereduced("unit", { unit })
		)
		if (branches.length === 1) {
			return branches[0]
		}
		return this.parseRoot("union", {
			branches
		})
	}
	parseTypeNode(schema, opts = {}) {
		const kind = assertTypeKind(schema)
		if (opts.allowedKinds && !opts.allowedKinds.includes(kind)) {
			return throwParseError(
				`Schema of kind ${kind} should be one of ${opts.allowedKinds}`
			)
		}
		return opts.root
			? this.parseRoot(kind, schema, opts)
			: this.parse(kind, schema, opts)
	}
	parseRoot(kind, def, opts = {}) {
		const node = this.parse(kind, def, opts)
		if (this.resolved) {
			// this node was not part of the original scope, so compile an anonymous scope
			// including only its references
			this.bindCompiledScope(node.contributesReferences)
		} else {
			// we're still parsing the scope itself, so defer compilation but
			// add the node as a reference
			Object.assign(
				this.referencesByName,
				node.contributesReferencesByName
			)
		}
		return node
	}
	parsePrereduced(kind, def) {
		return this.parse(kind, def, { prereduced: true })
	}
	parse(kind, def, opts = {}) {
		if (opts.alias && opts.alias in this.resolutions) {
			return throwInternalError(
				`Unexpected attempt to recreate existing alias ${opts.alias}`
			)
		}
		const node = parseAttachments(kind, def, {
			...opts,
			$: this,
			definition: def,
			prereduced: opts.prereduced ?? false
		})
		return node
	}
	bindCompiledScope(references) {
		const compiledTraversals = this.compileScope(references)
		for (const node of references) {
			if (node.jit) {
				// if node has already been bound to another scope or anonymous type, don't rebind it
				continue
			}
			node.jit = true
			node.traverseAllows =
				compiledTraversals[`${node.name}Allows`].bind(
					compiledTraversals
				)
			if (node.isType() && !node.includesContextDependentPredicate) {
				// if the reference doesn't require context, we can assign over
				// it directly to avoid having to initialize it
				node.allows = node.traverseAllows
			}
			node.traverseApply =
				compiledTraversals[`${node.name}Apply`].bind(compiledTraversals)
		}
	}
	compileScope(references) {
		return new CompiledFunction()
			.block(`return`, (js) => {
				references.forEach((node) => {
					const allowsCompiler = new NodeCompiler("Allows").indent()
					node.compile(allowsCompiler)
					const applyCompiler = new NodeCompiler("Apply").indent()
					node.compile(applyCompiler)
					js.line(
						allowsCompiler.writeMethod(`${node.name}Allows`) +
							",\n" +
							applyCompiler.writeMethod(`${node.name}Apply`) +
							","
					)
				})
				return js
			})
			.compile()()
	}
	schema = Object.assign(this.parseBranches.bind(this), {
		units: this.parseUnits.bind(this),
		union: this.parseUnion.bind(this)
	})
}
export const scopeNode = ScopeNode.from
export const rootSchema = (...branches) =>
	ScopeNode.root.parseTypeNode(
		branches.length === 1 ? branches[0] : branches,
		{
			root: true,
			prereduced: true
		}
	)
//# sourceMappingURL=schemaScope.js.map
