import { ReadonlyArray, hasDefinedKey } from "@arktype/util"
import { arkKind } from "../util.js"
import { pathToPropString } from "./utils.js"
export const throwArkError = (...args) => {
	throw new ArkError(...args)
}
export class ArkError extends TypeError {
	toString() {
		return this.message
	}
	throw() {
		throw this
	}
}
export const ArkTypeError = class extends ArkError {
	[arkKind] = "error"
	constructor(context) {
		super(context.message)
		Object.assign(this, context)
	}
}
export class ArkErrors extends ReadonlyArray {
	ctx
	constructor(ctx) {
		super()
		this.ctx = ctx
	}
	byPath = {}
	count = 0
	mutable = this
	add(error) {
		const pathKey = error.path.join(".")
		const existing = this.byPath[pathKey]
		if (existing) {
			const errorIntersection = createError(this.ctx, {
				code: "intersection",
				errors:
					existing.code === "intersection"
						? [...existing.errors, error]
						: [existing, error]
			})
			const existingIndex = this.indexOf(existing)
			// If existing is found (which it always should be unless this was externally mutated),
			// replace it with the new problem intersection. In case it isn't for whatever reason,
			// just append the intersection.
			this.mutable[existingIndex === -1 ? this.length : existingIndex] =
				errorIntersection
			this.byPath[pathKey] = errorIntersection
		} else {
			this.byPath[pathKey] = error
			this.mutable.push(error)
		}
		this.count++
	}
	get summary() {
		return this.toString()
	}
	toString() {
		return this.join("\n")
	}
	throw() {
		throw new ArkError(`${this}`, { cause: this })
	}
}
export const createError = (ctx, input) => {
	let errCtx
	const data = ctx.data
	const nodeConfig = ctx.config.predicate
	if (typeof input === "string") {
		errCtx = {
			code: "predicate",
			path: [...ctx.path],
			propString: pathToPropString(ctx.path),
			data,
			actual: nodeConfig.actual(data),
			expected: input
		}
		errCtx.problem = nodeConfig.problem(errCtx)
		errCtx.message = nodeConfig.message(errCtx)
	} else {
		const code = input.code ?? "predicate"
		const nodeConfig = ctx.config[code]
		const expected = input.expected ?? nodeConfig.expected?.(input)
		const path = input.path ?? [...ctx.path]
		errCtx = {
			...input,
			// prioritize these over the raw user provided values so we can
			// check for keys with values like undefined
			code,
			path,
			propString: pathToPropString(path),
			data: "data" in input ? input.data : data,
			actual:
				input.actual !== undefined
					? input.actual
					: nodeConfig.actual?.(data),
			expected
		}
		errCtx.problem = hasDefinedKey(input, "problem")
			? input.problem
			: nodeConfig.problem(errCtx)
		errCtx.message = hasDefinedKey(input, "message")
			? input.message
			: nodeConfig.message(errCtx)
	}
	return new ArkTypeError(errCtx)
}
//# sourceMappingURL=errors.js.map
