import {
	type ErrorMessage,
	type JsonData,
	type evaluate,
	type indexOf,
	type listable,
	type requireKeys
} from "@arktype/util"
import type { Node, UnknownNode } from "../base.js"
import type { Declaration, Inner, errorContext } from "../kinds.js"
import type { SchemaParseContext } from "../parse.js"
import type { NodeConfig, ParsedUnknownNodeConfig, Scope } from "../scope.js"
import type { Type, typeKindOrRightOf, typeKindRightOf } from "../types/type.js"
import type {
	BaseErrorContext,
	BaseMeta,
	BaseNodeDeclaration
} from "./declare.js"
import type { Disjoint } from "./disjoint.js"
export declare const basisKinds: readonly ["unit", "proto", "domain"]
export type BasisKind = (typeof basisKinds)[number]
export declare const propKinds: readonly ["prop", "index", "sequence"]
export type PropKind = (typeof propKinds)[number]
export declare const rangeKinds: readonly [
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after"
]
export type RangeKind = (typeof rangeKinds)[number]
export declare const boundKinds: readonly [
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after"
]
export type BoundKind = (typeof boundKinds)[number]
export declare const refinementKinds: readonly [
	"regex",
	"divisor",
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after"
]
export type RefinementKind = (typeof refinementKinds)[number]
export declare const constraintKinds: readonly [
	"regex",
	"divisor",
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after",
	"prop",
	"index",
	"sequence",
	"predicate"
]
export type ConstraintKind = (typeof constraintKinds)[number]
export declare const typeKinds: readonly [
	"union",
	"morph",
	"unit",
	"intersection",
	"proto",
	"domain"
]
export type TypeKind = (typeof typeKinds)[number]
export declare const intersectionChildKinds: readonly [
	"proto",
	"domain",
	"regex",
	"divisor",
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after",
	"prop",
	"index",
	"sequence",
	"predicate"
]
export type IntersectionChildKind = (typeof intersectionChildKinds)[number]
export type NodeKind = TypeKind | ConstraintKind
export declare const nodeKinds: [
	"union",
	"morph",
	"unit",
	"intersection",
	"proto",
	"domain",
	"regex",
	"divisor",
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after",
	"prop",
	"index",
	"sequence",
	"predicate"
]
export type OpenNodeKind = {
	[k in NodeKind]: Declaration<k>["intersectionIsOpen"] extends true
		? k
		: never
}[NodeKind]
export type ClosedNodeKind = Exclude<NodeKind, OpenNodeKind>
export declare const primitiveKinds: readonly [
	"unit",
	"proto",
	"domain",
	"regex",
	"divisor",
	"exactLength",
	"max",
	"min",
	"maxLength",
	"minLength",
	"before",
	"after",
	"predicate"
]
export type PrimitiveKind = (typeof primitiveKinds)[number]
export type CompositeKind = Exclude<NodeKind, PrimitiveKind>
export type OrderedNodeKinds = typeof nodeKinds
export declare const constraintKeys: {
	regex: 1
	divisor: 1
	exactLength: 1
	max: 1
	min: 1
	maxLength: 1
	minLength: 1
	before: 1
	after: 1
	prop: 1
	index: 1
	sequence: 1
	predicate: 1
}
export declare const propKeys: {
	prop: 1
	index: 1
	sequence: 1
	onExtraneousKey: 1
}
export declare const discriminatingIntersectionKeys: {
	readonly onExtraneousKey: 1
	readonly regex: 1
	readonly divisor: 1
	readonly exactLength: 1
	readonly max: 1
	readonly min: 1
	readonly maxLength: 1
	readonly minLength: 1
	readonly before: 1
	readonly after: 1
	readonly prop: 1
	readonly index: 1
	readonly sequence: 1
	readonly predicate: 1
}
type RightsByKind = accumulateRightKinds<OrderedNodeKinds, {}>
export type kindOrRightOf<kind extends NodeKind> = kind | kindRightOf<kind>
export type kindLeftOf<kind extends NodeKind> = Exclude<
	NodeKind,
	kindOrRightOf<kind>
>
export type kindOrLeftOf<kind extends NodeKind> = kind | kindLeftOf<kind>
type accumulateRightKinds<
	remaining extends readonly NodeKind[],
	result
> = remaining extends readonly [
	infer head extends NodeKind,
	...infer tail extends NodeKind[]
]
	? accumulateRightKinds<
			tail,
			result & {
				[k in head]: tail[number]
			}
		>
	: result
export type ConstraintIntersection<
	lKind extends ConstraintKind,
	rKind extends kindOrRightOf<lKind>
> = (l: Node<lKind>, r: Node<rKind>, $: Scope) => UnknownNode | Disjoint | null
export type ConstraintIntersectionMap<kind extends ConstraintKind> = evaluate<
	{
		[_ in kind]: ConstraintIntersection<kind, kind>
	} & {
		[rKind in kindRightOf<kind>]?: ConstraintIntersection<kind, rKind>
	}
>
export type TypeIntersection<
	lKind extends TypeKind,
	rKind extends typeKindOrRightOf<lKind>
> = (l: Node<lKind>, r: Node<rKind>, $: Scope) => Type | Disjoint
export type TypeIntersectionMap<kind extends TypeKind> = {
	[rKind in typeKindOrRightOf<kind>]: TypeIntersection<kind, rKind>
}
export type IntersectionMap<kind extends NodeKind> = kind extends TypeKind
	? TypeIntersectionMap<kind>
	: ConstraintIntersectionMap<kind & ConstraintKind>
export type UnknownIntersectionMap = {
	[k in NodeKind]?: (
		l: UnknownNode,
		r: UnknownNode,
		$: Scope
	) => UnknownIntersectionResult
}
export type UnknownIntersectionResult = UnknownNode | Disjoint | null
type PrecedenceByKind = {
	[i in indexOf<OrderedNodeKinds> as OrderedNodeKinds[i]]: i
}
export declare const precedenceByKind: {
	union: 0
	morph: 1
	unit: 2
	intersection: 3
	proto: 4
	domain: 5
	regex: 6
	divisor: 7
	exactLength: 8
	max: 9
	min: 10
	maxLength: 11
	minLength: 12
	before: 13
	after: 14
	prop: 15
	index: 16
	sequence: 17
	predicate: 18
}
export declare const isNodeKind: (value: unknown) => value is NodeKind
export declare function assertNodeKind<kind extends NodeKind>(
	value: unknown,
	kind: kind
): asserts value is Node<kind>
export type precedenceOfKind<kind extends NodeKind> = PrecedenceByKind[kind]
export declare const precedenceOfKind: <kind extends NodeKind>(
	kind: kind
) => precedenceOfKind<kind>
export type kindRightOf<kind extends NodeKind> = RightsByKind[kind]
export declare const typeKindsRightOf: <
	kind extends
		| "union"
		| "morph"
		| "unit"
		| "intersection"
		| "proto"
		| "domain"
>(
	kind: kind
) => typeKindRightOf<kind>[]
export type KeyDefinitions<d extends BaseNodeDeclaration> = {
	[k in keyRequiringDefinition<d>]: NodeKeyImplementation<d, k>
}
type keyRequiringDefinition<d extends BaseNodeDeclaration> = Exclude<
	keyof d["normalizedSchema"],
	keyof BaseMeta
>
export declare const defaultValueSerializer: (v: unknown) => JsonData
export type NodeKeyImplementation<
	d extends BaseNodeDeclaration,
	k extends keyof d["normalizedSchema"],
	instantiated = k extends keyof d["inner"] ? d["inner"][k] : never
> = requireKeys<
	{
		preserveUndefined?: true
		meta?: true
		child?: true
		implied?: true
		serialize?: (
			schema: instantiated extends listable<UnknownNode> | undefined
				? ErrorMessage<`Keys with node children cannot specify a custom serializer`>
				: instantiated
		) => JsonData
		parse?: (
			schema: Exclude<d["normalizedSchema"][k], undefined>,
			ctx: SchemaParseContext
		) => instantiated
	},
	| (d["normalizedSchema"][k] extends instantiated ? never : "parse")
	| ([instantiated] extends [listable<UnknownNode> | undefined]
			? "child"
			: never)
>
interface CommonNodeImplementationInput<d extends BaseNodeDeclaration> {
	keys: KeyDefinitions<d>
	normalize: (schema: d["schema"]) => d["normalizedSchema"]
	hasAssociatedError: d["errorContext"] extends null ? false : true
	collapsibleKey?: keyof d["inner"]
	reduce?: (
		inner: d["inner"],
		$: Scope
	) => Node<d["reducibleTo"]> | Disjoint | undefined
}
export interface UnknownNodeImplementation
	extends CommonNodeImplementationInput<BaseNodeDeclaration> {
	defaults: ParsedUnknownNodeConfig
	intersectionIsOpen: boolean
	intersections: UnknownIntersectionMap
	keys: Record<string, NodeKeyImplementation<any, any>>
}
export type nodeImplementationOf<d extends BaseNodeDeclaration> =
	nodeImplementationInputOf<d> & {
		intersections: IntersectionMap<d["kind"]>
		intersectionIsOpen: d["intersectionIsOpen"]
		defaults: nodeDefaultsImplementationFor<d["kind"]>
	}
export type nodeImplementationInputOf<d extends BaseNodeDeclaration> =
	CommonNodeImplementationInput<d> & {
		intersections: IntersectionMap<d["kind"]>
		defaults: nodeDefaultsImplementationInputFor<d["kind"]>
	} & (d["intersectionIsOpen"] extends true
			? {
					intersectionIsOpen: true
				}
			: {}) &
		(d["reducibleTo"] extends d["kind"]
			? {}
			: {
					reduce: {}
				})
type nodeDefaultsImplementationInputFor<kind extends NodeKind> = requireKeys<
	NodeConfig<kind>,
	| "description"
	| (Inner<kind> extends Omit<
			errorContext<kind>,
			keyof BaseErrorContext | "description"
	  >
			? never
			: "expected" & keyof NodeConfig<kind>)
>
export type nodeDefaultsImplementationFor<kind extends NodeKind> = Required<
	NodeConfig<kind>
>
export type DescriptionWriter<kind extends NodeKind = NodeKind> = (
	node: Node<kind>
) => string
export {}
//# sourceMappingURL=implement.d.ts.map
