import {
	compileSerializedValue,
	entriesOf,
	isKeyOf,
	throwInternalError
} from "@arktype/util"
import { Disjoint } from "../shared/disjoint.js"
const discriminantKinds = {
	domain: 1,
	value: 1
}
const parseDiscriminantKey = (key) => {
	const lastPathIndex = key.lastIndexOf("]")
	return [
		JSON.parse(key.slice(0, lastPathIndex + 1)),
		key.slice(lastPathIndex + 1)
	]
}
const discriminantCache = new Map()
export const discriminate = (branches) => {
	if (branches.length < 2) {
		return null
	}
	const cached = discriminantCache.get(branches)
	if (cached !== undefined) {
		return cached
	}
	// const pureValueBranches = branches.flatMap((branch) =>
	// 	branch.unit ? branch.unit : []
	// )
	// if (pureValueBranches.length === branches.length) {
	// 	const cases: DiscriminatedCases = transform(
	// 		pureValueBranches,
	// 		([i, valueNode]) => [valueNode.serialized, [branches[i]]]
	// 	)
	// 	return {
	// 		path: [],
	// 		kind: "value",
	// 		cases,
	// 		isPureRootLiteral: true
	// 	}
	// }
	const casesBySpecifier = {}
	for (let lIndex = 0; lIndex < branches.length - 1; lIndex++) {
		const l = branches[lIndex]
		for (let rIndex = lIndex + 1; rIndex < branches.length; rIndex++) {
			const r = branches[rIndex]
			const result = l.intersect(r)
			if (!(result instanceof Disjoint)) {
				continue
			}
			for (const { path, kind, disjoint } of result.flat) {
				if (!isKeyOf(kind, discriminantKinds)) {
					continue
				}
				const qualifiedDiscriminant = `${path}${kind}`
				let lSerialized
				let rSerialized
				if (kind === "domain") {
					lSerialized = disjoint.l
					rSerialized = disjoint.r
				} else if (kind === "value") {
					lSerialized = compileSerializedValue(disjoint.l)
					rSerialized = compileSerializedValue(disjoint.r)
				} else {
					return throwInternalError(
						`Unexpected attempt to discriminate disjoint kind '${kind}'`
					)
				}
				if (!casesBySpecifier[qualifiedDiscriminant]) {
					casesBySpecifier[qualifiedDiscriminant] = {
						[lSerialized]: [l],
						[rSerialized]: [r]
					}
					continue
				}
				const cases = casesBySpecifier[qualifiedDiscriminant]
				if (!isKeyOf(lSerialized, cases)) {
					cases[lSerialized] = [l]
				} else if (!cases[lSerialized].includes(l)) {
					cases[lSerialized].push(l)
				}
				if (!isKeyOf(rSerialized, cases)) {
					cases[rSerialized] = [r]
				} else if (!cases[rSerialized].includes(r)) {
					cases[rSerialized].push(r)
				}
			}
		}
	}
	// TODO: determinstic? Update cache key?
	const bestDiscriminantEntry = entriesOf(casesBySpecifier)
		.sort((a, b) => Object.keys(a[1]).length - Object.keys(b[1]).length)
		.at(-1)
	if (!bestDiscriminantEntry) {
		discriminantCache.set(branches, null)
		return null
	}
	const [specifier, predicateCases] = bestDiscriminantEntry
	const [path, kind] = parseDiscriminantKey(specifier)
	const cases = {}
	for (const k in predicateCases) {
		const subdiscriminant = discriminate(predicateCases[k])
		cases[k] = subdiscriminant ?? predicateCases[k]
	}
	const discriminant = {
		kind,
		path,
		cases,
		isPureRootLiteral: false
	}
	discriminantCache.set(branches, discriminant)
	return discriminant
}
// TODO: if deeply includes morphs?
export const writeUndiscriminableMorphUnionMessage = (path) =>
	`${
		path === "/" ? "A" : `At ${path}, a`
	} union including one or more morphs must be discriminable`
//# sourceMappingURL=discriminate.js.map
