import {
	append,
	appendUnique,
	conflatenateAll,
	entriesOf,
	isArray,
	isEmptyObject,
	omit,
	pick,
	splitByKeys,
	throwInternalError
} from "@arktype/util"
import { BaseNode } from "../base.js"
import { PropsGroup } from "../constraints/props/props.js"
import { metaKeys } from "../shared/declare.js"
import { Disjoint } from "../shared/disjoint.js"
import { constraintKeys, propKeys } from "../shared/implement.js"
import { BaseType, defineRightwardIntersections } from "./type.js"
const intersectionChildKeyParser = (kind) => (input, ctx) => {
	if (isArray(input)) {
		if (input.length === 0) {
			// Omit empty lists as input
			return
		}
		return input
			.map((schema) => ctx.$.node(kind, schema))
			.sort((l, r) => (l.innerId < r.innerId ? -1 : 1))
	}
	const node = ctx.$.node(kind, input)
	return node.intersectionIsOpen ? [node] : node
}
// 	readonly literalKeys = this.named.map((prop) => prop.key.name)
// 	readonly namedKeyOf = cached(() => node.unit(...this.literalKeys))
// 	readonly indexedKeyOf = cached(
// 		() =>
// 			new Type(
// 				this.indexed.flatMap((entry) => entry.key.branches),
// 				this.meta
// 			)
// 	)
// 	readonly keyof = cached(() => this.namedKeyOf().or(this.indexedKeyOf()))
// get(key: string | Type) {
// 	return typeof key === "string"
// 		? this.named.find((entry) => entry.value.branches)?.value
// 		: this.indexed.find((entry) => entry.key.equals(key))?.value
// }
const intersectIntersections = (reduced, raw, $) => {
	// avoid treating adding instance keys as keys of lRoot, rRoot
	if (reduced instanceof IntersectionNode) reduced = reduced.inner
	if (raw instanceof IntersectionNode) raw = raw.inner
	const [reducedConstraintsInner, reducedRoot] = splitByKeys(
		reduced,
		constraintKeys
	)
	const [rawConstraintsInner, rawRoot] = splitByKeys(raw, constraintKeys)
	// since intersection with a left operand of unknown is leveraged for
	// reduction, check for the case where r is empty so we can preserve
	// metadata and save some time
	const root = isEmptyObject(reduced)
		? rawRoot
		: intersectRootKeys(reducedRoot, rawRoot)
	if (root instanceof Disjoint) return root
	const lConstraints = flattenConstraints(reducedConstraintsInner)
	const rConstraints = flattenConstraints(rawConstraintsInner)
	return intersectConstraints({
		root,
		l: lConstraints,
		r: rConstraints,
		types: [],
		$
	})
}
export class IntersectionNode extends BaseType {
	static implementation = this.implement({
		hasAssociatedError: true,
		normalize: (schema) => schema,
		keys: {
			domain: {
				child: true,
				parse: intersectionChildKeyParser("domain")
			},
			proto: {
				child: true,
				parse: intersectionChildKeyParser("proto")
			},
			divisor: {
				child: true,
				parse: intersectionChildKeyParser("divisor")
			},
			max: {
				child: true,
				parse: intersectionChildKeyParser("max")
			},
			min: {
				child: true,
				parse: intersectionChildKeyParser("min")
			},
			maxLength: {
				child: true,
				parse: intersectionChildKeyParser("maxLength")
			},
			minLength: {
				child: true,
				parse: intersectionChildKeyParser("minLength")
			},
			exactLength: {
				child: true,
				parse: intersectionChildKeyParser("exactLength")
			},
			before: {
				child: true,
				parse: intersectionChildKeyParser("before")
			},
			after: {
				child: true,
				parse: intersectionChildKeyParser("after")
			},
			regex: {
				child: true,
				parse: intersectionChildKeyParser("regex")
			},
			predicate: {
				child: true,
				parse: intersectionChildKeyParser("predicate")
			},
			prop: {
				child: true,
				parse: intersectionChildKeyParser("prop")
			},
			index: {
				child: true,
				parse: intersectionChildKeyParser("index")
			},
			sequence: {
				child: true,
				parse: intersectionChildKeyParser("sequence")
			},
			onExtraneousKey: {
				parse: (def) => (def === "ignore" ? undefined : def)
			}
		},
		// leverage reduction logic from intersection and identity to ensure initial
		// parse result is reduced
		reduce: (inner, $) =>
			// we cast union out of the result here since that only occurs when intersecting two sequences
			// that cannot occur when reducing a single intersection schema using unknown
			intersectIntersections({}, inner, $),
		defaults: {
			description(node) {
				return node.children.length === 0
					? "unknown"
					: node.props?.description ??
							node.children
								.map((child) => child.description)
								.join(" and ")
			},
			expected(source) {
				return (
					"  • " + source.errors.map((e) => e.expected).join("\n  • ")
				)
			},
			problem(ctx) {
				return `must be...\n${ctx.expected}`
			}
		},
		intersections: {
			intersection: intersectIntersections,
			...defineRightwardIntersections("intersection", (l, r, $) => {
				// if l is unknown, return r
				if (l.children.length === 0) return r
				const basis = l.basis?.intersect(r) ?? r
				return basis instanceof Disjoint
					? basis
					: l?.basis?.equals(basis)
						? // if the basis doesn't change, return the original intesection
							l
						: // given we've already precluded l being unknown, the result must
							// be an intersection with the new basis result integrated
							$.node(
								"intersection",
								Object.assign(omit(l.inner, metaKeys), {
									[basis.kind]: basis
								}),
								{ prereduced: true }
							)
			})
		}
	})
	basis = this.domain ?? this.proto
	refinements = this.children.filter((node) => node.isRefinement())
	props = maybeCreatePropsGroup(this.inner, this.$)
	traversables = conflatenateAll(
		this.basis,
		this.refinements,
		this.props,
		this.predicate
	)
	expression =
		this.props?.expression ||
		this.children.map((node) => node.nestableExpression).join(" & ") ||
		"unknown"
	traverseAllows = (data, ctx) =>
		this.traversables.every((traversable) =>
			traversable.traverseAllows(data, ctx)
		)
	traverseApply = (data, ctx) => {
		if (this.basis) {
			this.basis.traverseApply(data, ctx)
			if (ctx.hasError()) return
		}
		if (this.refinements.length) {
			for (let i = 0; i < this.refinements.length - 1; i++) {
				this.refinements[i].traverseApply(data, ctx)
				if (ctx.failFast && ctx.hasError()) return
			}
			this.refinements.at(-1).traverseApply(data, ctx)
			if (ctx.hasError()) return
		}
		if (this.props) {
			this.props.traverseApply(data, ctx)
			if (ctx.hasError()) return
		}
		if (this.predicate) {
			for (let i = 0; i < this.predicate.length - 1; i++) {
				this.predicate[i].traverseApply(data, ctx)
				if (ctx.failFast && ctx.hasError()) return
			}
			this.predicate.at(-1).traverseApply(data, ctx)
		}
	}
	compile(js) {
		if (js.traversalKind === "Allows") {
			this.traversables.forEach((traversable) =>
				traversable instanceof BaseNode
					? js.check(traversable)
					: traversable.compile(js)
			)
			js.return(true)
			return
		}
		const returnIfFail = () => js.if("ctx.hasError()", () => js.return())
		const returnIfFailFast = () =>
			js.if("ctx.failFast && ctx.hasError()", () => js.return())
		if (this.basis) {
			js.check(this.basis)
			// we only have to return conditionally if this is not the last check
			if (this.traversables.length > 1) returnIfFail()
		}
		if (this.refinements.length) {
			for (let i = 0; i < this.refinements.length - 1; i++) {
				js.check(this.refinements[i])
				returnIfFailFast()
			}
			js.check(this.refinements.at(-1))
			if (this.props || this.predicate) returnIfFail()
		}
		if (this.props) {
			this.props.compile(js)
			if (this.predicate) returnIfFail()
		}
		if (this.predicate) {
			for (let i = 0; i < this.predicate.length - 1; i++) {
				js.check(this.predicate[i])
				// since predicates can be chained, we have to fail immediately
				// if one fails
				returnIfFail()
			}
			js.check(this.predicate.at(-1))
		}
	}
	rawKeyOf() {
		return this.basis
			? this.props
				? this.basis.rawKeyOf().or(this.props.rawKeyOf())
				: this.basis.rawKeyOf()
			: this.props?.rawKeyOf() ?? this.$.keywords.never
	}
}
const maybeCreatePropsGroup = (inner, $) => {
	const propsInput = pick(inner, propKeys)
	return isEmptyObject(propsInput) ? undefined : new PropsGroup(propsInput, $)
}
const intersectRootKeys = (l, r) => {
	const result = {}
	const lBasis = l.proto ?? l.domain
	const rBasis = r.proto ?? r.domain
	const resultBasis = lBasis
		? rBasis
			? lBasis.intersect(rBasis)
			: lBasis
		: rBasis
	if (resultBasis) {
		if (resultBasis instanceof Disjoint) {
			return resultBasis
		}
		if (resultBasis.kind === "domain" || resultBasis.kind === "proto") {
			result[resultBasis.kind] = resultBasis
		} else {
			return throwInternalError(
				`Unexpected intersection basis intersection ${resultBasis}`
			)
		}
	}
	if (l.onExtraneousKey || r.onExtraneousKey) {
		result.onExtraneousKey =
			l.onExtraneousKey === "throw" || r.onExtraneousKey === "throw"
				? "throw"
				: "prune"
	}
	return result
}
const intersectConstraints = (s) => {
	if (!s.r.length) {
		let result = s.$.node(
			"intersection",
			Object.assign(s.root, unflattenConstraints(s.l)),
			{ prereduced: true }
		)
		for (const type of s.types) {
			if (result instanceof Disjoint) {
				return result
			}
			result = type.intersect(result)
		}
		return result
	}
	const head = s.r.shift()
	let matched = false
	for (let i = 0; i < s.l.length; i++) {
		const result = s.l[i].intersect(head)
		if (result === null) continue
		if (result instanceof Disjoint) return result
		if (!matched) {
			if (result.isType()) s.types.push(result)
			else s.l[i] = result
			matched = true
		} else if (!s.l.includes(result)) {
			return throwInternalError(
				`Unexpectedly encountered multiple distinct intersection results for refinement ${result}`
			)
		}
	}
	if (!matched) {
		s.l.push(head)
	}
	head.impliedSiblings?.forEach((node) => appendUnique(s.r, node))
	return intersectConstraints(s)
}
const flattenConstraints = (inner) => {
	const result = entriesOf(inner)
		.flatMap(([k, v]) => (k in constraintKeys ? v : []))
		.sort((l, r) =>
			l.precedence < r.precedence
				? -1
				: l.precedence > r.precedence
					? 1
					: l.innerId < r.innerId
						? -1
						: 1
		)
	return result
}
const unflattenConstraints = (constraints) => {
	const inner = {}
	for (const constraint of constraints) {
		if (constraint.intersectionIsOpen) {
			inner[constraint.kind] = append(inner[constraint.kind], constraint)
		} else {
			if (inner[constraint.kind]) {
				return throwInternalError(
					`Unexpected intersection of closed refinements of kind ${constraint.kind}`
				)
			}
			inner[constraint.kind] = constraint
		}
	}
	return inner
}
//# sourceMappingURL=intersection.js.map
