import { listFrom, reference, throwParseError } from "@arktype/util"
import { Disjoint } from "../shared/disjoint.js"
import { basisKinds } from "../shared/implement.js"
import { BaseType, defineRightwardIntersections } from "./type.js"
export const morphChildKinds = ["intersection", ...basisKinds]
export class MorphNode extends BaseType {
	// TODO: recursively extract in?
	static implementation = this.implement({
		hasAssociatedError: false,
		keys: {
			in: {
				child: true,
				parse: (schema, ctx) =>
					ctx.$.node(schema, { allowedKinds: morphChildKinds })
			},
			out: {
				child: true,
				parse: (schema, ctx) =>
					ctx.$.node(schema, { allowedKinds: morphChildKinds })
			},
			morphs: {
				parse: listFrom,
				serialize: (morphs) => morphs.map(reference)
			}
		},
		normalize: (schema) => schema,
		defaults: {
			description(node) {
				return `a morph from ${node.in.description} to ${node.out.description}`
			}
		},
		intersections: {
			morph: (l, r, $) => {
				if (l.morphs.some((morph, i) => morph !== r.morphs[i])) {
					// TODO: is this always a parse error? what about for union reduction etc.
					// TODO: check in for union reduction
					return throwParseError(`Invalid intersection of morphs`)
				}
				const inTersection = l.in.intersect(r.in)
				if (inTersection instanceof Disjoint) {
					return inTersection
				}
				const outTersection = l.out.intersect(r.out)
				if (outTersection instanceof Disjoint) {
					return outTersection
				}
				return $.node("morph", {
					morphs: l.morphs,
					in: inTersection,
					out: outTersection
				})
			},
			...defineRightwardIntersections("morph", (l, r, $) => {
				const inTersection = l.in.intersect(r)
				return inTersection instanceof Disjoint
					? inTersection
					: inTersection.kind === "union"
						? $.node(
								"union",
								inTersection.branches.map((branch) => ({
									...l.inner,
									in: branch
								}))
							)
						: $.node("morph", {
								...l.inner,
								in: inTersection
							})
			})
		}
	})
	expression = `(In: ${this.in.expression}) => Out<${this.out.expression}>`
	serializedMorphs = this.morphs.map((morph) => reference(morph))
	traverseAllows = (data, ctx) => this.in.traverseAllows(data, ctx)
	traverseApply = (data, ctx) => {
		this.morphs.forEach((morph) => ctx.queueMorph(morph))
		this.in.traverseApply(data, ctx)
	}
	compile(js) {
		if (js.traversalKind === "Allows") {
			js.return(js.invoke(this.in))
			return
		}
		this.serializedMorphs.forEach((name) =>
			js.line(`ctx.queueMorph(${name})`)
		)
		js.line(js.invoke(this.in))
	}
	get in() {
		return this.inner.in
	}
	get out() {
		return this.inner.out ?? this.$.keywords.unknown
	}
	rawKeyOf() {
		return this.in.rawKeyOf()
	}
}
//# sourceMappingURL=morph.js.map
